//
//		cImage.h
//
// Alex Meyer -- a.meyer@@cs.ucl.ac.uk
//

#ifndef IMAGE_H
#define IMAGE_H

#include <fstream>
#include <iostream>
#include <ostream>
#include <string>
#include <stdio.h>
#include <assert.h>

#ifdef _GZLIB_
#include <zlib.h>
#endif

class cImage
{
public:
	cImage()															{ mX=0; mY=0; mC=0; mIm=NULL; misNew=false; }
	cImage(int a, int b, int d)									{ mX=a; mY=b; mC=d; mIm = newbuffer(mX*mY*mC); misNew=true; }
	cImage(int a, int b, int d, unsigned char* aim)		{ mX=a; mY=b; mC=d; mIm=aim; misNew=false; }
	inline cImage(std::string , int=-1);
	void initSize(int a, int b, int d)						{ if (misNew) freebuffer(); mX=a; mY=b; mC=d; mIm = newbuffer(mX*mY*mC); misNew=true; }
	~cImage()															{ if (misNew) freebuffer(); }
	inline void set(int dimx, int dimy, int col);
	void init(unsigned char uc=0)								{ int s=size(); assert(s>0); for(int i=0;i<s;i++) mIm[i]=uc; /*memset(im,uc,mX*mY*mC);*/ }
	int getWidth() const											{ return mX; }
	int getHeight() const										{ return mY; }
	int getDepth() const											{ return mC; }
	int getDimX() const											{ return mX; }
	int getDimY() const											{ return mY; }
	int getDimC() const											{ return mC; }
	void insert(const cImage& , int , int);
	void setRGBAforGLBlend();
	void merge(const cImage& mIm, int r, int g, int b, int a);
	unsigned char* getData() const							{ return mIm; }
	unsigned char* getData(int i, int j, int k)			{ return &(mIm[(j*mX+i)*mC+k]); }
	inline int savePPM(std::string , int bin=1, int version=-1) const;
	inline void loadPPM(std::string name);
	inline int halfSize(cImage* , int , int ) const;
	inline void blurTextureWithAlpha(int n=1);
	unsigned char& operator()(int i, int j, int k)		{ return mIm[ (j*mX+i)*mC + k ]; }
	unsigned char operator()(int i, int j, int k) const{ return mIm[ (j*mX+i)*mC + k ]; }
	int& operator()(int i, int j)							{ assert(mC==4); return *((int*)(mIm+((j*mX+i)*mC))); }
	int operator()(int i, int j) const					{ assert(mC==4); return *((int*)(mIm+((j*mX+i)*mC))); }
	int size() const												{ return mX*mY*mC; }
	inline void permutateX();
	inline void permutateY();
	inline void permutateXY();
	int readBuffer(std::ifstream& f)								{ f.read( ((char*)(mIm)), mX*mY*mC); return(!f.fail()); }
	void background(double r=1.0, double g=1.0, double b=1.0);
	void divideByAlpha();
	inline void getMinMaxFloat(float& min, float& max) const;
	inline void copy( cImage& src, int xdest, int ydest, int sizex=-1, int sizey=-1);
	inline bool isInit() const									{ return(mIm); }
private:
	int mX, mY, mC;
	unsigned char *mIm;
	bool misNew;
	inline unsigned char* newbuffer(int s);
	inline void freebuffer();
};

inline std::ostream& operator<<(std::ostream& o, const cImage& im)
{
	o<<"cImage "<<im.getDimX()<<"x"<<im.getDimY()<<" "<<im.getDimC()<<" depth color";
	return o;
}



template<class A> A Clamp(A a, A min=0.0, A max=1.0)
{
	if (a<min) return min;
	if (a>max) return max;
	return a;
}


inline unsigned char* cImage::newbuffer(int s)
{
#ifdef CFGSYS_SCIFS
	if (scifsunit==NULL)
		return ((unsigned char*)(new unsigned char[s]));
	else
		return ((unsigned char*)(scifsunit->scifs_malloc(s)));
#else
	return new unsigned char[s];
#endif
}



inline void cImage::set(int dimx, int dimy, int col)
{
	if (misNew) freebuffer(); 
	mX=dimx; 
	mY=dimy; 
	mC=col; 
	mIm = newbuffer(mX*mY*mC);
	misNew=true;
}



inline void cImage::permutateX()
{
	for(int i=0;i<mX/2;i++)
		for(int j=0;j<mY;j++)
			for(int k=0;k<mC;k++)
			{
				unsigned char tmp = operator()(i,j,k);
				operator()(i,j,k) = operator()(mX-1-i,j,k);
				operator()(mX-1-i,j,k) = tmp;
			}
}



inline void cImage::permutateY()
{
	for(int i=0;i<mX;i++)
		for(int j=0;j<mY/2;j++)
			for(int k=0;k<mC;k++)
			{
				unsigned char tmp = operator()(i,j,k);
				operator()(i,j,k) = operator()(i,mY-1-j,k);
				operator()(i,mY-1-j,k) = tmp;
			}
}



inline void cImage::permutateXY()
{
	for(int i=0;i<mX;i++)
		for(int j=0;j<mY/2;j++)
			for(int k=0;k<mC;k++)
			{
				unsigned char tmp = operator()(i,j,k);
				operator()(i,j,k) = operator()( mX-1-i, mY-1-j, k);
				operator()( mX-1-i, mY-1-j, k) = tmp;
			}
}



inline void cImage::freebuffer()
{
	delete[] mIm;
	mIm = NULL;
}



inline int cImage::halfSize(cImage* d, int dx, int dy) const
{
	//if ((d->getDimX()*dx!=getDimX()) || (d->getDimY()*dy!=getDimY())) return 1;
	int i,j,k,l,ic;
	float v,n;
	for(i=0;i<d->getDimX();i++)
		for(j=0;j<d->getDimY();j++)
			for(ic=0;ic<getDepth();ic++)
			{
				v=0.0;
				n=0.0;
				for(k=0;k<dx;k++)
					for(l=0;l<dy;l++)
					{
						v += static_cast<float>(operator()(i*dx+k,j*dy+l,ic));
						n+=1.0;
					}
				d->operator()(i,j,ic) = static_cast<unsigned char>(v/n);
			}
	return 0;
}



//! Blur the color when alpha=0 because GL interpolate with pixel around and this give black contour
inline void cImage::blurTextureWithAlpha(int np)
{
	if (mC!=4) return;
	int i,j,k,n,l,itmp,jtmp;
	for(n=0;n<np;n++)
	{
		for(i=0;i<getDimX();i++)
			for(j=0;j<getDimY();j++)
			{
				itmp = i;
				jtmp = j;
				switch(n%2) 
				{
					case 1 :
						i = getDimX()-1-i;
						j = getDimY()-1-j;
						break;
					case 2 :
						i = getDimX()-1-i;
						break;
					case 3 :
						j = getDimY()-1-j;
						break;
				}
				if (operator()(i,j,3)<=0)
				{
					float r = 0.0, g = 0.0, b = 0.0;
					float nv= 0.0;
					for(k=-2;k<=2;k++)
					{
						if ((i+k>=0) && (i+k<getDimX()))
							for(l=-2;l<=2;l++)
								if ((j+l>=0) && (j+l<getDimY()))
									//if (operator()(i,j,3)>5)
									if ((operator()(i+k,j+l,0)>2) || (operator()(i+k,j+l,1)>2) || 
										 (operator()(i+k,j+l,2)>2) || (operator()(i+k,j+l,3)>2))
									{
										r += operator()(i+k,j+l,0);
										g += operator()(i+k,j+l,1);
										b += operator()(i+k,j+l,2);
										nv += 1.0;
									}
					}
					if (nv>0.0)
					{
						operator()(i,j,0) = static_cast<unsigned char>(r/nv);
						operator()(i,j,1) = static_cast<unsigned char>(g/nv);
						operator()(i,j,2) = static_cast<unsigned char>(b/nv);
					}
				}
				i = itmp;
				j = jtmp;
			}
	}
}



inline cImage::cImage(std::string name, int nC)
{
	mIm = NULL;
	mX = 0;
	mY = 0;
	mC = nC;
	loadPPM(name);
}



inline void cImage::loadPPM(std::string name)
{
	using namespace std;

	if (name=="") return;

#ifdef _GZLIB_
	gzFile f;
#else
	FILE* f;
#endif
	
	int ul;
	unsigned char uc;
	int V;

#ifdef _GZLIB_
	f = gzopen(name.c_str(),"rb");
	if (f==NULL) 
	{
		cerr<<"cImage::loadPPM(std::string, int ): can not open "<<name<<endl;
		if (name.rfind(".gz")==string::npos)
			name += ".gz";
		else
			name = string( name, 0, name.rfind(".gz"));
			
		f = gzopen(name.c_str(),"rb");
	}
#else
	if (name.rfind(".gz")!=string::npos) 
	{ 
		std::cerr<<"cImage::loadPPM(std::string, int ) Compiled without _GZLIB_. Can not open "<<name;
		name = string( name, 0, name.rfind(".gz"));	
		std::cerr<<" I try "<<name<<endl;
	}
	f = fopen( name.c_str(), "rb");
#endif
	
	if (f==NULL) 
	{ 
		std::cerr<<"cImage::loadPPM(std::string, int ): can not open "<<name<<endl;
		return;
	}

	char line[256];
	int ax, ay;

#ifdef _GZLIB_
	gzgets( f, line, 256);
#else
	fgets( line, 256, f);
#endif
	sscanf( line, "P%d\n", &V);

	int NBCOL_real=0;
	switch(V)
	{
		case 7 :
			NBCOL_real = 4;
			break;
		case 6 : 
			NBCOL_real = 3;
			break;
		case 5 :
			NBCOL_real = 1;
			break;
		default :
			std::cerr<<"cImage::loadPPM(std::string, int) : "<<name<<" Bad version P"<<V<<" (try P5 or P6 or P7)!"<<endl;
			return;
	}

	do 
#ifdef _GZLIB_
		gzgets( f, line, 256);
#else
	fgets( line, 256, f);
#endif
	while( line[0]=='#' );
	
	sscanf( line, "%d %d", &ax, &ay);
	
	do 
#ifdef _GZLIB_
		gzgets( f, line, 256); 
#else
		fgets( line, 256, f); 
#endif
	while( line[0]=='#' );
	
	sscanf( line, "%ul", &ul);


	if (mIm==NULL)
	{
		mX = ax;
		mY = ay;
		mC = NBCOL_real;
		mIm = newbuffer(mX*mY*mC);
	}


	if ( (mX!=ax) || (mY!=ay) )
	{
		std::cerr<<"cImage::loadPPM : "<<name<<" Error size!"<<endl;
		return;
	}
	
	init(255);
	int m = (mC>NBCOL_real)?mC:NBCOL_real;
	for(int j=0;j<mY;j++)
		for(int i=0;i<mX;i++)
			for(int k=0;k<m;k++)
			{
				if (k<NBCOL_real) 
#ifdef _GZLIB_
					gzread( f, &uc, 1); 
#else
					fread( &uc, sizeof(uc), 1, f);
#endif
				else 
					uc=0;
				if (k<mC) operator()(i,j,k) = uc;
			}
#ifdef _GZLIB_
	gzclose(f);
#else
	fclose(f);
#endif
}



inline int cImage::savePPM(std::string name, int bin, int version) const
{
	using namespace std;

#ifdef _GZLIB_
	gzFile f;
#else
	FILE* f;
#endif


#ifdef _GZLIB_
	f = gzopen(name.c_str(),"wb");
	if (name.rfind(".gz")==string::npos)
		name += ".gz";
#else
	f = fopen( name.c_str(), "wb");
#endif


	
	if (f==NULL) { std::cerr<<"cImage::savePPM : can't open "<<name<<endl; return 0; }

	if (version<=0)
		if (getDepth()<3) version=5;
		else version=6;

//	unsigned char uc;
	switch(version)
	{
		case 7 :
			if (getDepth()<4) std::cerr<<"cImage::savePPM(): depth of cImage does not match P7 version!"<<endl;
#ifdef _GZLIB_
			gzprintf( f, "P7\n%d %d\n255\n", mX,mY);
#else
			fprintf( f, "P7\n%d %d\n255\n", mX,mY);
#endif
			if (bin)
			{
				for(int j=0;j<mY;j++)
					for(int i=0;i<mX;i++)
#ifdef _GZLIB_
						gzwrite( f, mIm+(j*mX+i)*mC, 4);
#else
						fwrite( mIm+(j*mX+i)*mC, sizeof(unsigned char), 4, f);
#endif
			}
			else std::cerr<<"cImage::savePPM() : not implemented!(use bin)"<<endl;
			break;
		case 6 :
			if (getDepth()<3) std::cerr<<"cImage::savePPM(): depth of cImage does not match P7 version!"<<endl;
#ifdef _GZLIB_
			gzprintf( f, "P6\n%d %d\n255\n", mX,mY);
#else
			fprintf( f, "P6\n%d %d\n255\n", mX,mY);
#endif
			if (bin)
			{
				for(int j=0;j<mY;j++)
					for(int i=0;i<mX;i++)
#ifdef _GZLIB_
						gzwrite( f, mIm+(j*mX+i)*mC, 3);
#else
						fwrite( mIm+(j*mX+i)*mC, sizeof(unsigned char), 3, f);
#endif
			}
			else std::cerr<<"cImage::savePPM() : not implemented!(use bin)"<<endl;
			break;
		case 5 :
#ifdef _GZLIB_
			gzprintf( f, "P5\n%d %d\n255\n", mX,mY);
#else
			fprintf( f, "P5\n%d %d\n255\n", mX,mY);
#endif
			if (bin)
			{
				for(int j=0;j<mY;j++)
					for(int i=0;i<mX;i++)
#ifdef _GZLIB_
						gzwrite( f, mIm+(j*mX+i)*mC+mC-1, 1);
#else
						fwrite( mIm+(j*mX+i)*mC+mC-1, sizeof(unsigned char), 1, f);
#endif

			}
			else std::cerr<<"cImage::savePPM() : not implemented!(use bin)"<<endl;
			break;
		default:
			std::cerr<<"cImage::savePPM(std::string) : P"<<version<<" not implemented! Sorry!"<<endl;
			break;
	}
	
#ifdef _GZLIB_
	gzclose(f);
#else
	fclose(f);
#endif
	return(1);
}


//! Compute (RGBA)=(R*alpha, G*alpha, B*alpha, alpha)
inline void cImage::setRGBAforGLBlend()
{
	if (getDepth()!=4) return;
	int i,j;
	double alpha;
	for(i=0;i<getDimX();i++)
		for(j=0;j<getDimY();j++)
		{
			alpha = static_cast<double>(operator()(i,j,3))/255.0;
			operator()(i,j,0) *= static_cast<unsigned char>(alpha);
			operator()(i,j,1) *= static_cast<unsigned char>(alpha);
			operator()(i,j,2) *= static_cast<unsigned char>(alpha);
		}
}


//! Merge this and im -> this with parameters r=0|1,g=0|1,b=0|1,a=0|1|2
inline void cImage::merge(const cImage& im, int r, int g, int b, int a)
{
	if ((getDepth()!=4) || (im.getDepth()!=4) || (im.getDimX()!=getDimX()) || (im.getDimY()!=getDimY())) return;
	int i,j;
	for(i=0;i<getDimX();i++)
		for(j=0;j<getDimY();j++)
		{
			if (r!=0) operator()(i,j,0) = im(i,j,0);
			if (g!=0) operator()(i,j,1) = im(i,j,1);
			if (b!=0) operator()(i,j,2) = im(i,j,2);
			if (a!=0) operator()(i,j,3) = im(i,j,3);
			if (a==2) operator()(i,j,3) = 255-operator()(i,j,3);
		}
}



inline void cImage::insert(const cImage& cImage, int px, int py)
{
	assert(cImage.getDepth()==getDepth());
	for(int i=0;i<cImage.getWidth();i++)
		for(int j=0;j<cImage.getHeight();j++)
			for(int k=0;k<cImage.getDepth();k++)
			{
				assert( (px+i<getWidth()) && (py+j<getHeight()) );
				(*this)(px+i,py+j,k) = cImage(i,j,k);
			}
}



inline void cImage::background(double r, double g, double b)
{
	assert(getDepth()==4);
	double bg[3];
	bg[0]=r;
	bg[1]=g;
	bg[2]=b;
	
	for(int i=0;i<getWidth();i++)
		for(int j=0;j<getHeight();j++)
		{
			if (operator()(i,j,3)==0)
			{
				operator()(i,j,0) = 0;
				operator()(i,j,1) = 0;
				operator()(i,j,2) = 0;
			}

			double alpha = ((double)(operator()(i,j,3)))/(255.0);
			for(int k=0;k<3;k++)
			{
				double mC = ((double)(operator()(i,j,k)))/255.0;
				double v = alpha*mC + (1.0-alpha)*bg[k];
				//double v = mC + (1.0-alpha)*bg[k];
			
				operator()(i,j,k) = ((unsigned char)(Clamp(255.0*v,0.0,255.0)));
			}
		}
}



inline void cImage::divideByAlpha()
{
	assert(getDepth()==4);
	
	for(int i=0;i<getWidth();i++)
		for(int j=0;j<getHeight();j++)
		{
			if (operator()(i,j,3)!=0)
			{
				double alpha = ((double)(operator()(i,j,3)))/(255.0);
				for(int k=0;k<3;k++)
				{
					double mC = ((double)(operator()(i,j,k)))/255.0;
					double v = mC/alpha;
			
					operator()(i,j,k) = ((unsigned char)(Clamp(255.0*v,0.0,255.0)));
				}
			}
		}
}



inline void cImage::getMinMaxFloat(float& min, float& max) const
{
	int size = mX*mY;
	int i;
	float* f=((float*)(mIm));
	min = f[0];
	max = f[0];
	for(i=0;i<size;i++)
	{
		if (f[i]<min) min=f[i];
		if (f[i]>max) max=f[i];
	}
}



inline void cImage::copy( cImage& src, int xdest, int ydest, int sizex, int sizey)
{
	assert(src.getDepth()==getDepth());

	int sx = (sizex<0)?src.getWidth():sizex;
	int sy = (sizey<0)?src.getWidth():sizey;
	
	if (xdest+sx > getWidth() )  sx = getWidth()-xdest;
	if (ydest+sy > getHeight() ) sy = getHeight()-ydest;
	
	int i,j,k;
	for(i=0;i<sx;i++)
		for(j=0;j<sy;j++)
			for(k=0;k<getDepth();k++)
			{
				operator()(xdest+i,ydest+j, k) = src(i,j,k);
			}
}

#endif
